package slowlogs

import (
	"context"
	"net/http"
	"net/http/httptest"
	"testing"
	"time"

	"github.com/sirupsen/logrus"
	testlog "github.com/sirupsen/logrus/hooks/test"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"gitlab.com/gitlab-org/labkit/log"
)

// loggerFunc will inject request host and path to the logged messages
func loggerFunc(r *http.Request) *logrus.Entry {
	return log.WithFields(log.Fields{
		"host": r.Host,
		"path": r.URL.Path,
	})
}

func TestMiddlewareLogsSlowRequests(t *testing.T) {
	slowHandler := http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		time.Sleep(600 * time.Millisecond) // Simulate slow processing time
		w.WriteHeader(http.StatusOK)
	})

	hook := testlog.NewGlobal()
	middleware := NewMiddleware(slowHandler, loggerFunc)

	req := httptest.NewRequest(http.MethodGet, "/", nil)
	rr := httptest.NewRecorder()
	middleware.ServeHTTP(rr, req)

	assert.Equal(t, http.StatusOK, rr.Code, "Expected HTTP status OK")

	lastEntry := hook.LastEntry()
	require.NotNil(t, lastEntry, "Expected a log entry for slow requests")

	require.Contains(t, lastEntry.Data, "middleware_time_distribution", "Expected 'execution_time_distribution' in log entry")
	require.Contains(t, lastEntry.Data, "total_execution_time_ms", "Expected 'total_execution_time_ms' in log entry")

	totalExecutionTime := lastEntry.Data["total_execution_time_ms"].(int64)
	assert.GreaterOrEqual(t, totalExecutionTime, int64(600), "Expected 'total_execution_time_ms' to be greater than or equal to 600 milliseconds")
}

func TestMiddlewareDoesNotLogFastRequests(t *testing.T) {
	fastHandler := http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		time.Sleep(200 * time.Millisecond) // Simulate fast processing time
		w.WriteHeader(http.StatusOK)
	})

	hook := testlog.NewGlobal()
	middleware := NewMiddleware(fastHandler, loggerFunc)

	req := httptest.NewRequest(http.MethodGet, "/", nil)
	rr := httptest.NewRecorder()

	middleware.ServeHTTP(rr, req)

	assert.Equal(t, http.StatusOK, rr.Code)

	lastEntry := hook.LastEntry()
	require.Nil(t, lastEntry, "Log entry is not expected")
}

func TestPrepareMiddlewareTimingInfoInMS(t *testing.T) {
	timings := map[string]time.Duration{
		"handler1": 150 * time.Millisecond,
		"handler2": 200 * time.Millisecond,
	}

	ctx := context.WithValue(context.Background(), timingContextKey, &timingInfo{
		MiddlewareTimings: timings,
	})

	timingsInMS := prepareMiddlewareTimingInfoInMS(ctx)

	require.NotNil(t, timingsInMS)
	assert.Equal(t, int64(150), timingsInMS["handler1_ms"])
	assert.Equal(t, int64(200), timingsInMS["handler2_ms"])
}

// TestLogHandlerTimingWithChainedHandlers tests LogHandlerTiming with handler chaining.
func TestLogHandlerTimingWithChainedHandlers(t *testing.T) {
	slowHandler := func() http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			time.Sleep(300 * time.Millisecond)
			w.WriteHeader(http.StatusOK)
		})
	}

	anotherSlowHandler := func(handler http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			time.Sleep(400 * time.Millisecond)
			handler.ServeHTTP(w, r)
		})
	}

	hook := testlog.NewGlobal()

	chainedHandler := LogHandlerTiming(slowHandler(), "slowHandler")
	chainedHandler = LogHandlerTiming(anotherSlowHandler(chainedHandler), "anotherSlowHandler")
	chainedHandler = NewMiddleware(chainedHandler, loggerFunc)

	req := httptest.NewRequest(http.MethodGet, "/", nil)
	rr := httptest.NewRecorder()

	chainedHandler.ServeHTTP(rr, req)

	assert.Equal(t, http.StatusOK, rr.Code)

	require.NotNil(t, hook.LastEntry())
	require.NotNil(t, hook.LastEntry().Data["middleware_time_distribution"])

	executionTimes := hook.LastEntry().Data["middleware_time_distribution"].(map[string]int64)
	handler1Duration, ok := executionTimes["slowHandler_ms"]
	require.True(t, ok)
	handler2Duration, ok := executionTimes["anotherSlowHandler_ms"]
	require.True(t, ok)
	totalExecutionTimeMS, ok := hook.LastEntry().Data["total_execution_time_ms"].(int64)
	require.True(t, ok)

	// Assert that the first handler execution time is approximately 300 milliseconds
	assert.GreaterOrEqual(t, handler1Duration, int64(300))
	assert.Less(t, handler1Duration, int64(350)) // Allow some margin

	// Assert that the second handler execution time is approximately 400 milliseconds
	assert.GreaterOrEqual(t, handler2Duration, int64(400))
	assert.Less(t, handler2Duration, int64(450)) // Allow some margin

	assert.GreaterOrEqual(t, totalExecutionTimeMS, int64(700))
	assert.Less(t, totalExecutionTimeMS, int64(750)) // Allow some margin
}
